#!/usr/bin/env python
# encoding: utf-8
"""
manifest.py

Created by David Peckham on 2009-09-09.
Copyright (c) 2009 David Peckham. All rights reserved.
"""

import sys
import os
import unittest

from mako.template import Template


class Manifest:
    def __init__(self):
        pass

    def save(self, f, universe):
        """docstring for write_manifest"""
        template = Template(filename='templates/manifest.mako', 
            output_encoding='utf-8', encoding_errors='replace')
        f.write(template.render(universe=universe))

class ManifestTests(unittest.TestCase):
    def setUp(self):
        pass


if __name__ == '__main__':
    unittest.main()


#!/usr/bin/env python
# encoding: utf-8
"""
reader.py

Created by David Peckham on 2009-09-07.
Copyright (c) 2009 David Peckham. All rights reserved.
"""

import datetime
import os
import pdb
import repr
import struct
import sys

sys.path.insert(0, '..')
from pyunv.universe import Universe, Parameters, Class, Object, Condition, Table, VirtualTable, Column

# import pyunv

class Reader(object):
    
    _content_markers = ('Objects;', 'Tables;', 'Columns;', 
        'Virtual Tables;', 'Parameters;', 'Columns Id;')
    
    def __init__(self, f):
        super(Reader, self).__init__()
        self.file = f
        self.find_content_offsets()
        self.universe = Universe()
        self.universe.parameters = self.read_parameters()
        self.universe.tables = self.read_tables()
        self.universe.table_map = dict()
        for table in self.universe.tables:
            self.universe.table_map[table.id_] = table
        self.universe.virtual_tables = self.read_virtual_tables()
        self.universe.columns = self.read_columns()
        #self.universe.column_attributes = self.read_column_attributes()
        self.universe.joins = self.read_joins()
        self.universe.classes = self.read_classes()

    def find_content_offsets(self):
        """find the offsets of the object, table, and column definitions 
        in the BusinessObjects universe file"""
        self.content_offsets = dict()
        contents = self.file.read()
        for marker in Reader._content_markers:
            self.content_offsets[marker] = contents.find(chr(0)+marker) + len(marker) + 1
        del contents
        return
    
    def read_parameters(self):
        """docstring for read_parameters
        
        I unknown (usually 0x22)
        I unknown
        S universe_filename
        S universe_name
        I revision
        H unknown
        S description
        S created_by
        S modified_by
        I created_date
        I modified_date
        I query_time_limit (seconds)
        I row_limit
        S unknown
        S object_strategy
        x unknown
        I cost_estimate_warning_limit (seconds)
        I long_text_limit (characters)
        4x unknown
        S comments
        3I unknown
        S domain
        S dbms_engine
        S network_layer
        
         Other parameter blocks we don't parse yet:
            Parameters_4_1;
            Parameters_5_0;
            Parameters_6_0;
            Parameters_11_5;
        
        """
        self.file.seek(self.content_offsets['Parameters;'])
        
        params = Parameters()
        struct.unpack('<2I', self.file.read(8))
        params.universe_filename = self.read_shortstring()
        params.universe_name = self.read_shortstring()
        params.revision, = struct.unpack('<I', self.file.read(4))
        struct.unpack('<H', self.file.read(2))
        params.description = self.read_shortstring()
        params.created_by = self.read_shortstring()
        params.modified_by = self.read_shortstring()
        created, modified, = struct.unpack('<2I', self.file.read(8))
        params.created_date = self.date_from_dateindex(created)
        params.modified_date = self.date_from_dateindex(modified)
        seconds, = struct.unpack('<I', self.file.read(4))
        params.query_time_limit = seconds / 60
        params.row_limit, = struct.unpack('<I', self.file.read(4))
        self.read_shortstring()
        params.object_strategy = self.read_shortstring()
        struct.unpack('<x', self.file.read(1))
        seconds, = struct.unpack('<I', self.file.read(4))
        params.cost_estimate_warning_limit = seconds / 60
        params.long_text_limit, = struct.unpack('<I', self.file.read(4))
        struct.unpack('<4x', self.file.read(4))
        params.comments = self.read_shortstring()
        struct.unpack('<3I', self.file.read(12))
        params.domain = self.read_shortstring()
        params.dbms_engine = self.read_shortstring()
        params.network_layer = self.read_shortstring()
        
        return params

    def read_tables(self):
        """read a BusinessObjects schema definition from the universe file

        B unknown (usually 0x1)
        B unknown (usually 0x1 or 0x2)
        S schema_name
        S database_username
        I max_table_id
        I table_count
        ???B tables

        """
        self.file.seek(self.content_offsets['Tables;'])
        struct.unpack('<2B', self.file.read(2))
        schema_name = self.read_shortstring()
        user_name = self.read_shortstring()
        max_table_id, = struct.unpack('<I', self.file.read(4))
        table_count, = struct.unpack('<I', self.file.read(4))
        return [self.read_table() for x in range(table_count)]

    def read_virtual_tables(self):
        """read the virtual table definitions from the universe file

        I virtual_table_count
        ???B virtual_tables

        """
        self.file.seek(self.content_offsets['Virtual Tables;'])
        count, = struct.unpack('<I', self.file.read(4))
        return [self.read_virtualtable() for x in range(count)]

    def read_columns(self):
        """read the list of source database columns from the universe file

        I column_count
        I column_count?
        ???B columns

        """
        self.file.seek(self.content_offsets['Columns Id;'])
        column_count, = struct.unpack('<I', self.file.read(4))
        column_count2, = struct.unpack('<I', self.file.read(4))
        #print('count1 %d  count2 %d' % (column_count, column_count2))
        return [self.read_column() for x in range(column_count2)]
    
    # def read_column_attributes(self):
    #     """read the column attributes (after marker Columns;)"""
    #     pass

    def read_joins(self):
        """docstring for read_joins"""
        return []

    def read_classes(self):
        """docstring for read_classes"""
        self.file.seek(self.content_offsets['Objects;'])
        class_count, object_count, condition_count, rootclass_count, = \
            struct.unpack('<4I', self.file.read(16))
        return [self.read_class() for x in range(rootclass_count)]
        
    def read_table(self):
        """read a table definition from the universe file

        I table_id
        7x
        3I unknown
        S table_name
        13x
        ? flag
        H count
        xxI unknown (count times)
        
        """
        id_, = struct.unpack('<I', self.file.read(4))
        self.file.read(19)
        name = self.read_shortstring()
        self.file.read(13)
        flag, = struct.unpack('<?', self.file.read(1))
        if flag:
          count, = struct.unpack('<H', self.file.read(2))
          self.file.read(4*count+3)
        else:
          self.file.read(1)
        return Table(id_, name)

    def read_virtualtable(self):
        """read a virtual table definition from the universe file

        I table_id
        S select
        
        """
        table_id, = struct.unpack('<I', self.file.read(4))
        select = self.read_shortstring()
        return VirtualTable(table_id, select)

    def read_column(self):
        """read a column definition from the universe file

        I column_id
        I table_id
        S table_name
        
        """
        id_, = struct.unpack('<I', self.file.read(4))
        parent, = struct.unpack('<I', self.file.read(4))
        name = self.read_shortstring()
        #print(name)
        return Column(id_, name, parent, self.universe)

    def read_class(self):
        """read a BusinessObjects class definition from the universe file

        I subclass_count
        I id
        S name
        I parent
        S description
        7B unknown
        I object_count
        ???B objects
        I condition_count
        ???B conditions
        ???B subclasses

        """
        id_, = struct.unpack('<I', self.file.read(4))
        name = self.read_shortstring()
        parent, = struct.unpack('<I', self.file.read(4))
        description = self.read_shortstring()
        c = Class(self.universe, id_, parent, name, description)
        self.file.seek(7, os.SEEK_CUR)
        object_count, = struct.unpack('<I', self.file.read(4))
        c.objects = [self.read_object() for x in range(object_count)]
        condition_count, = struct.unpack('<I', self.file.read(4))
        c.conditions = [self.read_condition() for x in range(condition_count)]
        subclass_count, = struct.unpack('<I', self.file.read(4))
        c.subclasses = [self.read_class() for x in range(subclass_count)]
        return c

    def read_object(self):
        """read a BusinessObjects object definition from the universe file

        I id
        S name
        I parent
        S description
        H select_table_count
        ?I select_table_ids (repeats select_table_count times)
        H where_table_count
        ?I where_table_ids (repeats where_table_count times)
        S select (starts 03 nn* 2E)
        S where (starts 02 nn* 20)
        S format
        S unknown
        S lov_name
        58B unknown  (LOV settings, hide indicator?)

       """
        id_, = struct.unpack('<I', self.file.read(4))
        name = self.read_shortstring()
        parent, = struct.unpack('<I', self.file.read(4))
        description = self.read_shortstring()
        o = Object(self.universe, id_, parent, name, description)
        select_tablecount, = struct.unpack('<H', self.file.read(2))
        struct.unpack('<%dI' % select_tablecount, self.file.read(4 * select_tablecount))
        where_tablecount, = struct.unpack('<H', self.file.read(2))
        struct.unpack('<%dI' % where_tablecount, self.file.read(4 * where_tablecount))
        o.select = self.read_shortstring()
        o.where = self.read_shortstring()
        o.format = self.read_shortstring()
        unknown2 = self.read_shortstring()
        o.lov_name = self.read_shortstring()
        self.file.seek(58, os.SEEK_CUR)
        return o

    def read_condition(self):
        """read a BusinessObjects condition definition from the universe file

        I id
        S name
        I parent
        S description
        H where_tablecount
        ?I where_table_ids (repeats where_tablecount times)
        H unknown_tablecount
        ?I table_ids (repeats unknown_tablecount times)
        S where

        """
        id_, = struct.unpack('<I', self.file.read(4))
        name = self.read_shortstring()
        parent, = struct.unpack('<I', self.file.read(4))
        description = self.read_shortstring()
        c = Condition(self.universe, id_, name, parent, description)
        where_tablecount, = struct.unpack('<H', self.file.read(2))
        struct.unpack('<%dI' % where_tablecount, self.file.read(4 * where_tablecount))
        unknown_tablecount, = struct.unpack('<H', self.file.read(2))
        struct.unpack('<%dI' % unknown_tablecount, self.file.read(4 * unknown_tablecount))
        c.where = self.read_shortstring()
        return c

    def read_shortstring(self):
        """read a short variable-length string from the universe file"""
        length, = struct.unpack('<H', self.file.read(2))
        if length:
            s, = struct.unpack('<%ds' % length, self.file.read(length))
            return unicode(s.translate(None, '\x0d\x0a'), errors='ignore')
        else:
            return None

    def date_from_dateindex(self, dateindex):
        """return the date corresponding to the BusinessObjects universe date index"""
        #print(dateindex)
        return datetime.date(1976,7,4) + datetime.timedelta(dateindex-2442964)


import csv
import unittest
from manifest import Manifest

class ReaderTests(unittest.TestCase):
    def setUp(self):
        self.filename = 'tests/Universe_A8.unv'
            
    # def testexport(self):
    #     universe = Reader(open(self.filename, 'rb')).universe
    #     CsvWriter(universe, open(os.path.splitext(self.filename)[0]+'.csv', 'wb'))

    def testmanifest(self):
        universe = Reader(open(self.filename, 'rb')).universe
        f = open(self.filename+'.manifest.txt', 'w')
        manifest = Manifest()
        manifest.save(f, universe)
        f.close()
        

if __name__ == '__main__':
    unittest.main()





# BusinessObjects XI R2 constants

# I don't use these yet -- they're here for future reference

# alignment_type = { 'horizontal', 'indent', 'row_by_row_autofit', 'text_wrap', 'vertical'  }
cardinality = { 0:'Unknown', 1:'OneToOne', 2:'OneToMany', 3:'ManyToOne', 4:'ManyToMany' }
column_key = { 0:'Unknown', 1:'All', 2:'None', 3:'Primary', 4:'Secondary', 5:'Extern' }
column_type = { 0:'Null', 1:'Numeric', 2:'Character', 3:'Date', 4:'Text', 5:'Unknown' }
connection_state = { 1:'KeepActiveWholeSession', 2:'KeepActiveForxx', 3:'DisconnectAfterEachTransaction' }
connection_type = { 1:'Personal', 2:'Secured', 3:'Shared' }
object_aggregate = { 1:'Sum', 2:'Max', 3:'Min', 4:'Avg', 5:'Count', 6:'Null' }
object_qualification = { 1:'Dimension', 2:'Detail', 3:'Measure' }
object_security_access = { 0:'Public', 1:'Controlled', 2:'Restricted', 3:'Confidential', 4:'Private' }
object_type = { 0:'Null', 1:'Numeric', 2:'Character', 3:'Date', 4:'Blob', 5:'Unknown' }
outer_join = { 1:'None', 2:'Left', 3:'Right', 4:'Full' }
variable_type = { 1:'String', 2:'Numeric', 3:'Date' }

#
# attribute - name, parent, value (string)
#
# class - attributes, classes, description, name, objects, parent, predefined_conditions, root_class, show (boolean)
#
# column - key (column_key), name, parent, type (column_type)
#
# condition (aka predefined condition) - attributes, description, name, 
#    parent, root_class, show, tables, where
#
# connection - active_state (connection_state), active_time (long), 
#    array_bind_size (long), array_fetch_size (long), CUID (string), 
#    db_engine, db_source, is_async_mode, login_timeout (long), name, 
#    network_layer, password, perform_cost_estimate (boolean), 
#    server (string), type (connection_type), username, 
#    use_singlesignon (boolean)
#
# context - description, joins, name, parent
#
# control_option - cost_estimate_exceeded_value (long), limit_execution_time (bool),
#    limit_execution_time_value (long), limit_size_of_long_text_object (bool),
#    limit_size_of_long_text_object_value (long), limit_size_of_result_set (bool),
#    limit_size_of_result_set_value (long), parent, warn_if_cost_estimate_exceeded (bool)
#
# custom_hierarchy, default_hierarchy - dimensions, name, parent
#
# format - alignment (alignment_type), font, number_format (string), parent
#
# font - bold (bool), color (long), italic (bool), name, parent, size (long), 
#    strikethrough (bool), underline (bool)
#
# join - cardinality, expression, first_columns, first_table, outer_join, 
#        parent, second_columns, second_table, is_shortcut
#
# key - is_enabled, parent, select (string), type (column_key?)
#
# linked_universe - description, fullname, longname, name, parent
#
# list_of_values - name, parent, values
#
# object - active_month (bool), active_quarter (bool), active_year (bool), aggregate_function (object_aggregate), 
#    allow_user_to_edit_lov (bool), associated_dimension (object), attributes, 
#    autorefresh_lov_beforeuse (bool), can_be_used_condition (bool), can_be_used_result (bool), 
#    can_be_used_sort (bool), db_format (string), description, export_lov_with_universe (bool), 
#    format, has_lov, keys, list_of_values, month_name (string), name, objects, 
#    parent, qualification (object_qualification), quarter_name (string), root_class, 
#    security_access_level (object_security_access), select, show (bool), tables, type (object_type), 
#    use_hierarchical_display (bool), where, year_name (string)
#
# object_strategy - help (string), name, parent
#
# parameter - name, parent, value
#
# sql_option - complex_operators, multiple_sql_for_context, 
#    multiple_sql_for_measure, operators, parent, prevent_cartesion_products,
#    select_multiple_contexts, subqueries
#      (all except parent are booleans)
# 
# table - columns, incompatible_objects, incompatible_predefined_conditions,
#    is_alias, is_derived, name, original_table (table), parent, 
#    sql_of_derived_table (string), sql_of_derived_table_with_alias (string), 
#    weight (long), xpos (long), ypos (long)
#
# universe - author, candidate_classes, candidate_joins, classes, comments,
#    connection, contexts, control_option, creation_date, current_join_strategy,
#    current_object_strategy, current_owner, current_qualifier, 
#    current_table_strategy, custom_hierarchies, db_tables, default_hierarchies,
#    description, fullname, joins, join_strategies, language, linked_universes,
#    long_name, modification_date, modifier (string), name, object_strategies, owners,
#    owner_supported, parameters, parent, path, qualifiers, qualifier_supported,
#    saved, sql_option, tables, table_strategies, use_custom_hierarchies, windows
#
# variable - interpert_as (variable_type), is_multi_valued, name, parent, value (string)
#


#!/usr/bin/env python
# encoding: utf-8
"""
csvwriter.py

Created by David Peckham on 2009-09-07.
Copyright (c) 2009 David Peckham. All rights reserved.
"""

import sys
import os
import unittest
import csv

class CsvWriter(object):
    """Write an inventory of a universe's classes, objects, 
        and conditions to a CSV file

    class, name, id, parent, description, select, where

    """
    def __init__(self, universe, csvfile):
        super(CsvWriter, self).__init__()
        self.universe = universe
        self.file = csvfile
        writer = csv.writer(self.file, delimiter=',', quotechar='"', quoting=csv.QUOTE_MINIMAL)
        for c in universe.classes:
            self.write_class(writer, c, '')

    def write_class(self, writer, c, classpath):
        if classpath:
            classpath = classpath + '\\' + c.name
        else:
            classpath = c.name
        writer.writerow((classpath, None, 'class', c.description, None, None))
        for obj in c.objects:
            self.write_object(writer, obj, classpath)
        for condition in c.conditions:
            self.write_condition(writer, condition, classpath)
        for subclass in c.subclasses:
            self.write_class(writer, subclass, classpath)

    def write_object(self, writer, o, classpath):
        writer.writerow((classpath, o.name, 'object', o.description, o.select_sql, o.where_sql))

    def write_condition(self, writer, c, classpath):
        writer.writerow((classpath, c.name, 'condition', c.description, None, c.where_sql))



#!/usr/bin/env python
# encoding: utf-8
"""
pyunv.py

Created by David Peckham on 2009-09-01.
Copyright (c) 2009 David Peckham. All rights reserved.
"""

import os
import re
import sys


class Universe(object):
    
    """docstring for Universe"""
    def __init__(self, id_=None, name=None, description=None):
        super(Universe, self).__init__()
        self.id_ = id_
        self.name = name
        self.description = description
        self.tables = []
        self.columns = []
        self.classes = []
        self.joins = []
        self.parameters = None


class Parameters(object):
    """docstring for Parameters"""
    def __init__(self):
        super(Parameters, self).__init__()
        self.universe_filename = None
        self.universe_name = None
        self.revision = 0
        self.description = None
        self.created_by = None
        self.modified_by = None
        self.created_date = None
        self.modified_date = None
        self.query_time_limit = 0
        self.row_limit = 0
        self.object_strategy = None
        self.cost_estimate_warning_limit = 0
        self.long_text_limit = 0
        self.comments = None
        self.domain = None
        self.dbms_engine = None
        self.network_layer = None


class Class(object):
    
    """docstring for Class"""
    def __init__(self, universe, id_, parent, name, description):
        super(Class, self).__init__()
        self.universe = universe
        self.id_ = id_
        self.parent = parent
        self.name = name
        self.description = description
        self.objects = []
        self.conditions = []
        self.subclasses = []


class ObjectBase(object):
    """docstring for ObjectBase"""
    def __init__(self, universe, id_, parent, name, description):
        super(ObjectBase, self).__init__()
        assert(universe)
        self.universe = universe
        self.id_ = id_
        self.parent = parent
        self.name = name
        self.description = description
        self.select_table_refs = []
        self.where_table_refs = []
        self.select = None
        self.where = None

    def lookup_table(self, match):
        index = int(match.groups()[0])
        try:
            table = self.universe.table_map[index]
            return table.name
        except KeyError:
            return 'UNKNOWN (%d)' % index
    
    def expand_sql(self, sql):
        """Return the SQL with table names instead of table IDs"""
        if sql:
            p = re.compile(r'(?:'+chr(3)+')([0-9]{1,4})')
            return p.sub(self.lookup_table, sql)
        else:
            return None
        
    @property
    def select_sql(self):
        return self.expand_sql(self.select)

    @property
    def where_sql(self):
        return self.expand_sql(self.where)

    def __str__(self):
        return '%s id=%d, name=%s, select=%s' % (type(self), 
            self.id_, self.name, self.select) 

       
class Object(ObjectBase):
    
    """docstring for Object"""
    def __init__(self, universe, id_, parent, name, description):
        super(Object, self).__init__(universe, id_, parent, name, description)
        self.format = None
        self.lov_name = None
        

class Condition(ObjectBase):
    
    """docstring for Condition"""
    def __init__(self, universe, id_, parent, name, description):
        super(Condition, self).__init__(universe, id_, parent, name, description)


class Table(object):

    """docstring for Table"""
    def __init__(self, id_=None, name=None):
        super(Table, self).__init__()
        self.id_ = id_
        self.name = name

    def __str__(self):
        return '%s id=%d, name=%s' % (type(self), 
            self.id_, self.name) 
        

class VirtualTable(object):

    """docstring for VirtualTable"""
    def __init__(self, table_id=None, select=None):
        super(VirtualTable, self).__init__()
        self.table_id = table_id
        self.select = select

    def __str__(self):
        return '%s table_id=%d, select=%s' % (type(self), 
            self.table_id, self.select) 


class Column(object):

    """docstring for Column"""
    def __init__(self, id_=None, name=None, parent=None, universe=None):
        super(Column, self).__init__()
        self.id_ = id_
        self.name = name
        self.parent = parent
        self.universe = universe

    @property
    def table_name(self):
        try:
            tablename = self.universe.table_map[self.parent].name
        except KeyError:
            tablename = 'UNKNOWN TABLE (%d)' % self.parent
        return tablename

    def __str__(self):
        return '%s id=%d, name=%s parent=%d' % (type(self), 
            self.id_, self.name, self.parent) 
        




from setuptools import setup, find_packages
#from distutils.core import setup

# python setup.py register sdist upload


SUMMARY = """
PyUnv reads SAP BusinessObjects universe (.unv) files. PyUnv can extract 
universe settings, classes, objects, conditions, and source tables and columns
from the universe file.

PyUnv requires Mako to produce manifests. I include a sample template for a
text manifest. If you come up with your own manifests in RST, HTML, or other,
let me know.
"""

CLASSIFIERS = """
Development Status :: 2 - Pre-Alpha
Environment :: Other Environment
Intended Audience :: Developers
License :: OSI Approved :: GNU Library or Lesser General Public License (LGPL)
Natural Language :: English
Operating System :: OS Independent
Programming Language :: Python
Topic :: Software Development :: Libraries :: Python Modules
"""


setup(
    name = 'pyunv',
    version = '0.1.8',
    
    author='David Peckham',
    author_email = 'dave.peckham@me.com',
    classifiers = filter(None, CLASSIFIERS.split("\n")),
    description = 'Python parser for SAP BusinessObjects universe (*.unv) files',
    download_url = 'http://code.google.com/p/pyunv/downloads/list',
    install_requires = ['Mako'],
    keywords = ['encoding', 'BusinessObjects', 'SAP', 'universe', 'unv'],
    license = 'LGPL',
    long_description = SUMMARY.strip(),
    package_dir = {'':'lib'},
    packages = find_packages('lib', exclude=[]),
    platforms = ['Many'],
    provides = ['pyunv'],
    test_suite = 'pyunv.tests',
    url = 'http://code.google.com/p/pyunv/',
    zip_safe = False,
    )
